/*! \file layer.cpp
 * \brief Main library for the pagefault assisted automatic tuning.
 *
 * This is the main portion of the library which handles the automatic
 * tuning of MPI code. The library can be injected either dynamically
 * (via LD_PRELOAD) or statically (by including mpii.h and linking
 * with the generated injlib.a.
 *
 * Note that most variables, functions and classes in this file are
 * limited to the file-scope (static), to avoid polluting the
 * function namespace of the process we're injecting into, and
 * also allow the compiler as much flexibility with inlining as
 * it cares to do.
 *
 * \todo Use MPI_Irsend for MPI_Rsend. Or maybe not, as programs
 *       which use MPI_Rsend are kinda optimized already.
 * \todo If seen>10, make it a persistant request.
 * \todo Runtime option to mprotect() entire area of a chain.
 * \todo LD_PRELOAD oveeride for operator new.
 */

#include <mpi.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <limits.h>
#include <signal.h>
#include <sys/mman.h>
#include <errno.h>
#include <set>
#include <map>

using namespace std;

//! \brief Size of a hardware page, in bytes. 
#define PAGE_SIZE 16384

//! \brief And-mask for aligning an address to the start of page. 
#define PAGE_MASK 0xffffffffffffc000

//! \brief Minimum number of stable iterations before recognizing chain
#define CHAIN_MIN_ITER 3

// Make sure either INJECT_STATIC or INJECT_DYNAMIC is set, but not both
#ifndef INJECT_STATIC
#ifndef INJECT_DYNAMIC
#error Must set injection type
#endif
#else
#ifdef INJECT_DYNAMIC
#error Can not have both injection types
#endif
#endif

// If we're dynamically injecting, don't use a two-stage wrapper, just
// replace the function directly. We'll use dynamic loader tricks to
// find the right address.

#ifdef INJECT_DYNAMIC
#define Inj_MPI_Init MPI_Init
#define Inj_MPI_Send MPI_Send
#define Inj_MPI_Recv MPI_Recv
#define Inj_MPI_Sendrecv MPI_Sendrecv
#define Inj_MPI_Get_count MPI_Get_count
#define Inj_malloc malloc
#define Inj_free free
#endif

//! \brief Calling convention and name mangling of exported symbols.
#define INJ_METHOD extern "C"

//! \brief Holds the rank of the process that should output debug information.
static int do_debug;

//! \brief If nonzero, the user has requested no "messing" should take place.
static int do_nothing;

//! \brief Rank of current process. 
static int rank;

//! \brief Pointer to original MPI_Init() function.
static int (*Orig_MPI_Init) (int *, char ***);

//! \brief Pointer to original MPI_Isend() function.
static int (*Orig_MPI_Isend) (void *, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);

//! \brief Pointer to original MPI_Irecv() function.
static int (*Orig_MPI_Irecv) (void *, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);

//! \brief Pointer to original MPI_Wait() function.
static int (*Orig_MPI_Wait) (MPI_Request *, MPI_Status *);

//! \brief Pointer to original MPI_Get_count() function.
static int (*Orig_MPI_Get_count) (MPI_Status *, MPI_Datatype, int *);

//! \brief Pointer to original malloc() function.
static void *(*Orig_malloc)(size_t);

//! \brief Pointer to original free() function.
static void (*Orig_free)(void *);

//! \brief Information about new and old signal handler.
static struct sigaction sa;
static struct sigaction oldsa;

/*! \brief A range of memory.
 *
 * This is a convenience class to work with an manipulate a range
 * of memory (such as the buf variable of most MPI calls).
 * 
 * As many compilers don't like pointer arithmetics with void * pointers,
 * this makes the rest of the codemuch cleaner.
 */
struct Range {
	//! \brief Start of range.
	void *start;
	//! \brief End of range.
	void *stop;

	//! \brief Null constructor.
	Range() {
		start = stop = NULL;
	}
	
	/*! \brief Basic contructor.
	 * \param start Start of range.
	 * \param stop End of range.
         */

	Range(void *start, void *stop) {
		this->start = start;
		this->stop = stop;
	}

	/*! \brief Compare two ranges
 	 * \param o Other range to compare with.
	 * \return true if ranges are identical.
	 */
	bool operator ==(const Range &o) const {
		return ((o.start == start) && (o.stop==stop));
	}

	/*! \brief Check if two ranges overlap (intersect).
	 * \param o Other range to compare with.
	 * \return true if ranges overlap.
	 */
	bool overlaps(const Range &o) const {
		return ( ((o.start >= start) && (o.start <= stop)) ||
			 ((o.stop >= start) && (o.stop <= stop)));
	}

	/*! \brief Check if address is inside range.
	 * \param ptr Address to check.
	 * \return true if address is in range.
	 */
	bool overlaps(void *ptr) const {
		return ((ptr >= start) && (ptr <= stop));
	}

	/*! \brief Adjust range to page boundries.
	 * \return New adjusted range.
	 *
	 * start is adjusted downwards to the start of page boundry, and
	 * stop is likewise adjusted upwards.
	 */
	Range pageAdjust() const {
		Range o((void *)((long int)start & PAGE_MASK), (void *)(((long int)stop & PAGE_MASK) + (PAGE_SIZE - 1)));
		return o;
	}

	/*! \brief Compute size of range.
	 * \return Size of range in bytes.
	 */
	size_t size() const {
		return (long int)stop-(long int)start;
	}
};

/*! \brief A memory area in use by an active MPI request.
 *
 * This struct holds all the information about an active request.
 */

struct MemReq {
	//! \brief Indicates if is a MPI_*send or MPI_*recv request.
	bool is_send;

	//! \brief Is the memory area mprotect()ed?
	bool is_protected;
	/*! \brief Is this request the end of a chain?
	 *
 	 * If this is true, this is the end of a chain and we should
	 * wait for all outstanding requests to finish before 
	 * moving along.
	 */
	bool is_waitall;

	//! \brief Memory range (in user program) of request.
	Range reqmem;
	//! \brief Memory range of request, page-aligned for 
	// fast access.
	Range pages;

	//! \brief Base address used in request, identical to reqmem.start.
	void *base;
	//! \brief Number of elements in request.
	int count;
	//! \brief Datatype used in request.
	MPI_Datatype datatype;
	//! \brief Request handle.
	MPI_Request req;
	//! \brief Communicator used in request.
	MPI_Comm comm;
	//! \brief Status (if any)
	MPI_Status *status;

	//! \brief Size of buffer for MPI_Pack and MPI_Unpack
	int bufsize;
	//! \brief Allocated buffer for packing/unpacking data.
	unsigned char *buffer;

	/*! \brief Construct a new active request.
	 * \param is_send Is this request a MPI_*send?
	 * \param base Buffer to send/recv to.
	 * \param count Number of elements.
	 * \param datatype Datatype.
	 * \param comm Communicator to use.
	 * \param status Pointer to status variable.
	 * This will initialize reqmem and pages based on the datatype and
         * count, and will allocate a buffer for packing and unpacking.
	 */

	MemReq(bool is_send, unsigned char *base, int count, MPI_Datatype datatype, MPI_Comm comm, MPI_Status *status) {
		MPI_Aint extent, lb, ub;

		this->base = base;
		this->count = count;
		this->datatype = datatype;
		this->comm = comm;
		this->is_send = is_send;
		this->status = status;

		MPI_Type_lb(datatype, &lb);
		MPI_Type_ub(datatype, &ub);
		MPI_Type_extent(datatype, &extent);
		MPI_Pack_size(count, datatype, comm, &bufsize);

		MPI_Aint size = (count-1) * ub + extent;

		reqmem.start = base + lb;
		reqmem.stop = base + size;
		pages = reqmem.pageAdjust();

		buffer=new unsigned char[bufsize];

		is_protected = true;
		is_waitall = false;
	}

	/*! \brief Destructor.
	 *
	 * Deallocates memory for buffer.
	 */
	~MemReq() {
		delete [] buffer;
	}
};

/*! \brief Signature of MPI Request.
 *
 * This is the signature of an MPI request. If two signatures are identical,
 * they should transfer the same range of memory in the same way to the same
 * destination. This is used to recognize requests we've seen before on
 * subsequent iterations in the user program.
 *
 * The member variables here will be the parameters seen to the MPI_Send
 * or MPI_Recv calls.
 *
 * This struct has no virtual functions, no parent and should not have
 * any children as it relies on being "just a chunk of memory" for
 * its assignment and comparison operators.
 */

struct ReqSig {
	//! \brief Is this is a send request?
	bool is_send;
	//! \brief Base memory address of request.
	void *base;
	//! \brief Number of elements.
	int count;
	//! \brief Datatype 
	MPI_Datatype datatype;
	//! \brief Source or destination rank.
	int partner;
	//! \brief Message tag.
	int tag;
	//! \brief Communicator handle.
	MPI_Comm comm;

	/*! \brief Construct a new signature.
	 * \param is_send Is this a send requst?
	 * \param base The initial address of the send/receive buffer.
	 * \param count Number of elements in buffer.
	 * \param datatype Datatype of each element.
	 * \param partner Rank of source or destination.
	 * \param tag Message tag.
	 * \param comm Communicator handle.
	 *
	 * This constructor just does assignments, so it should be 
	 * reasonably fast. C++ will provide copy and assignment
	 * constructors for us (which will basically be memcpy()
	 * versions).
	 */	 
	ReqSig(bool is_send, void *base, int count, MPI_Datatype datatype, int partner, int tag, MPI_Comm comm) {
		this->is_send=is_send;
		this->base=base;
		this->count=count;
		this->datatype=datatype;
		this->partner=partner;
		this->tag=tag;
		this->comm=comm;
	}

	/*! \brief Compare two signatures.
	 * \param o Other signature to compare with.
	 * \return true if signatures are identical.
	 * \see operator<()
	 *
	 */
	bool operator ==(const ReqSig &o) const {
		return(memcmp(this, &o, sizeof(ReqSig))==0);
	}

	/*! \brief Compare two signatures.
	 * \param o Other signature to compare with.
	 * \return true if this signature is "less than" o.
	 * \see operator==()
	 *
	 * This is primarily to allow use as key in std::map<>.
	 */
	bool operator<(const ReqSig &o) const {
		return(memcmp(this, &o, sizeof(ReqSig))==-1);
	}
	
};

/*! \brief Information about a request we've seen.
 *
 * The library tracks and records the most recent requests done
 * by the user program. This is used to build information
 * about request chains; requests that always follow each other
 * without any computation inbetween them. Such chains can
 * then be started with their original buffer area, removing
 * the overhead of copying, and we simply wait at the end of
 * the chain for all requests to finish.
 *
 * Chains are recognized in add_req(), and are simply a series
 * of requests where prevcount and nextcount are fairly high,
 * and the chain terminates in the request which has faultat
 * set.
 */
struct ReqInfo {
	//! \brief How many times this exact request has been seen.
	int seen;
	//! \brief Next and previous request in chain.
	struct ReqInfo *prev, *next;
	//! \brief How many times have we seen the same next and previous requests.
	int prevcount, nextcount;
	//! \brief Address of page fault while this request was last active request.
	void *faultat;
	//! \brief If true, some earlier request specified this as chain end.
	bool waitall;
	ReqInfo() {
		seen=0;
		prev = next = NULL;
		prevcount=nextcount = 0;
		faultat = NULL;
		waitall = false;
	}
};

//! \brief Set of active requests.
static set<MemReq *> requests;

//! \brief Map of statuses we might wait for.
static map<MPI_Status *, MemReq *> statuses;

//! \brief Map between returned page-aligned address and "true" address.
static map<void *, void *> mallocmap;
//! \brief Map of malloc sizes (number of bytes)
static map<void *, size_t> mallocsizes;
//! \brief Set of pages we've pre-aligned (page number, not address).
static set<unsigned long int> mallocpages;

//! \brief Map of signatures and matching information.
static map<ReqSig, ReqInfo *> reqhist;

//! \brief Last active request.
static ReqInfo *last_req = NULL;

/*! \brief Output debug string
 * \param format format of string (printf format).
 *
 * This outputs a debug string to stderr if the environment 
 * variable INJ_DEBUG equals the rank of the calling process.
 */
 
static void dbgout(const char *format, ...)
{
	static int line = 1;

    if ((do_debug == rank) || (do_debug==-2)) {
	va_list args;


	fprintf(stderr, "%d[%d]: ", rank,line++);
	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end(args);
	fprintf(stderr, "\n");
    }
}

/*! \brief Wait for request and release memory.
 * \param mr Request to wait for.
 *
 * This will wait for an MPI request to finish, then
 * wait for any requests using the same pages it has
 * locked, and finally unprotect the pages so 
 * they can be accessed by the application again. 
 * If the request was a receive, the data will be
 * copied from the buffer to the user buffer.
 *
 * If the request was part of chain and not
 * protected, just wait for it finish.
 */

static void wait_req(MemReq *mr) {
	MPI_Status status;
	if (mr->status) 
		statuses.erase(mr->status);
	else
		mr->status = &status;

	dbgout("MPI_Wait %p .. status %p",mr,mr->status);
	MPI_Wait(&mr->req, mr->status);
	requests.erase(mr);

	// Memory ranges cannot overlap, but pages might, and we don't want
	// to unprotect the pages of requests that aren't done, so wait
	// for them too.

	set<MemReq *>::const_iterator iter;
	bool ok;

	do {
		ok = true;
		for(iter=requests.begin();iter != requests.end(); ++iter) {
			if ((*iter)->pages.overlaps(mr->pages) && (*iter)->is_protected) {
				dbgout("Found partner %p to wait for", *iter);
				wait_req(*iter);
				ok = false;
				break;
			}
		}
	} while (!ok);

	// Is this a unprotected request? (Part of chain)
	if (! mr->is_protected) {
		dbgout("Unprotected, early exit");
		delete mr;
		return;
	}
	
	// Get number of bytes transferred
	int bsize = 0;
	Orig_MPI_Get_count(mr->status, MPI_PACKED, &bsize);


	// Unprotect pages
	mprotect(mr->pages.start, mr->pages.size(), PROT_READ | PROT_WRITE);

	// Copy from buffer to userprogram if this was a receive.
	if (! mr->is_send) {
		dbgout("Going to copy");
		if (!bsize || bsize == MPI_UNDEFINED)
			return;
		int pos = 0;
		MPI_Unpack(mr->buffer, bsize, &pos, mr->base, mr->count, mr->datatype, mr->comm);
	}

	dbgout("Goint to delete and return");
	delete mr;
}

/*! \brief Wait for all outstanding requests.
 *
 * This waits for all outstanding requests, used at the end of chains and
 * in the case of "true" segment faults.
 */
void wait_all_req() {
	set<MemReq *>::iterator iter;

	dbgout("Wait all req");

	while((iter=requests.begin())!=requests.end()) 
		wait_req(*iter);
}

/*! \brief Start new request.
 * \param is_send Is this a send request?
 * \param base Base address of request.
 * \param count Number of elements.
 * \param datatype Datatype of elements.
 * \param rank Source or destination rank.
 * \param tag Message tag.
 * \param comm Communicator handle.
 * \param status Pointer to status receptor.
 * \param ignore Address of MemReq for which we ignore overlapping. (SendRecv pairs)
 * \return Address of MemReq
 *
 * This adds a new tracked request. First, it updates chain information, and 
 * if there is sufficient confidence this is part of a chain, skip
 * memory protection alltogether. If this is the end of a chain, wait
 * for all outstanding requests to finish before returning.
 *
 * For new requests, or requests not part of a chain, make sure any requests
 * which use the same memory addresses are already done (unless they were
 * a send, and this is also a send).
 */

static MemReq *add_req(bool is_send, void *base, int count, MPI_Datatype datatype, int rank, int tag, MPI_Comm comm, MPI_Status *status, MemReq *ignore = NULL) {
	if (rank == MPI_PROC_NULL)
		return NULL;

	ReqSig rs(is_send,base,count,datatype,rank,tag,comm);
	ReqInfo *ri;
	bool nopage;
	
	// Find previous ReqInfo for this signature, or make a new one.
	ri = reqhist[rs];
	if (! ri) {
		ri = new ReqInfo;
		reqhist[rs] = ri;
	}
	ri->seen++;

	// Was the previous request the same as last time?
	if (ri->prev == last_req) {
		ri->prevcount++;
	} else {
		if (ri->prev) {
			ri->prev->next = NULL;
			ri->prev->nextcount = 0;
		}
		ri->prev = last_req;
		ri->prevcount=1;
	}

	// Was this the "next" request of the previous
	// request lasttime?
	if (last_req) {
		if (last_req->next == ri) {
			last_req->nextcount++;
		} else {
			last_req->next = ri;
			last_req->nextcount++;
		}
	} else if (ri->prev) {
		ri->prev = NULL;
		ri->prevcount = 0;
	}

	last_req = ri;

	// Test confidence of chain. If it is a chain,
	// we have high nextcounts.
	ReqInfo *rptr = ri;
	int steps = 1;
	while (rptr->next && rptr->nextcount>=CHAIN_MIN_ITER) {
		rptr = rptr->next;
		steps++;
	}

	// Did we end up at something that pagefaulted?
	if (rptr && rptr->faultat && rptr != ri && steps>=2) {
		rptr->waitall = true;
		nopage = true;
	} else {
		nopage = false;
	}

	dbgout("ReqInfo %p: Next %p prev %p Ncnt %d Pcnt %d steps %d Waitall %d Npage %d",ri, ri->next, ri->prev, ri->nextcount, ri->prevcount, steps, ri->waitall,nopage);

	// Allocate new active request.
	MemReq *mr = new MemReq(is_send,(unsigned char *)base,count,datatype,comm,status);

	if (nopage || ri->waitall)
		mr->is_protected = false;
	if (ri->waitall)
		mr->is_waitall = true;

	bool memlap = false;

	// If memory ranges overlap, wait for request to finish
	set<MemReq *>::const_iterator iter;
	bool ok;
	do {
		ok = true;
		
		for(iter=requests.begin();ok && iter!=requests.end();++iter) {
			 if ((*iter)->pages.overlaps(mr->pages) && ((mr->is_protected || (*iter)->is_protected))) {
				if (mr->is_protected && (*iter)->is_protected)
					memlap = true;
				else if (*iter == ignore) {
					// This is a sendrecv pair and they disagree, which 
					// is troublesome. No choice but to equate them, possibly waiting.
					mr->is_protected = (*iter)->is_protected;
					if (! mr->is_protected)
						mr->is_waitall = true;
					ok = false;
					break;
				} else {
					// Overlapping and different types.
					wait_req(*iter);
					ok = false;
					break;
				}
			 }
		}
	} while (!ok);

	// If it is a send request with protected memory, copy into sending buffer
	int pos = 0;

	if (memlap && is_send)
		mprotect(mr->pages.start, mr->pages.size(), PROT_READ | PROT_WRITE);

	if (is_send && mr->is_protected)
			MPI_Pack(base, count, datatype, mr->buffer, mr->bufsize, &pos, comm);

	if (mr->is_protected) {
		// Mark pages
		if (mprotect(mr->pages.start, mr->pages.size(), PROT_NONE) != 0) {
			perror("MPII failed to access protect pages!");
			abort();
		}

		if (is_send)
			Orig_MPI_Isend(mr->buffer, pos, MPI_PACKED, rank, tag, comm, &mr->req);
		else
			Orig_MPI_Irecv(mr->buffer, mr->bufsize, MPI_PACKED, rank, tag, comm, &mr->req);
	} else {
		if (is_send)
			Orig_MPI_Isend(base, count, datatype, rank, tag, comm, &mr->req);
		else
			Orig_MPI_Irecv(base, count, datatype, rank, tag, comm, &mr->req);
	}

	requests.insert(mr);
	if (status)
		statuses[status] = mr;

	// If this was the end of a chain, wait for everything to be done.
	if (mr->is_waitall) {
		wait_all_req();
		last_req = NULL;
	}

	return mr;
}

/*! \brief Segment violation signal handler.
 * \param signal Signal number.
 * \param siginfo Information about signal.
 * \param ucontext User context.
 *
 * This handles page faults, which most likely are caused by the user
 * program accessing protected pages.
 *
 * We simply iterate the active requests, wait for all requests that
 * overlap the faulted address to finish. wait_req() will unprotect
 * the pages again, so we simply return to userspace when done.
 */

static void sigact(int signal, siginfo_t *siginfo, void *ucontext) {
	void *addr = siginfo->si_addr;

	bool found = false;

	dbgout("Sigaction");

	// If memory ranges overlap, wait for request to finish
	set<MemReq *>::const_iterator iter;
	bool ok;
	do {
		ok = true;
		for(iter=requests.begin();iter!=requests.end();++iter) {
			dbgout("Testing %p", *iter);
			if((*iter)->pages.overlaps(addr)) {
				found = true;
				ok = false;
				dbgout("Found. Waiting.");
				wait_req(*iter);
				break;
			}
		}
	} while (!ok);

	dbgout("Did iter");

	// If we didn't find anything, this is a "real" segment violation,
	// so we should abort.
	if (! found) {
		do_debug = rank;
		dbgout("MPII: SIGSEGV -- trying to chain\n");

		if (sigaction(SIGSEGV, &oldsa, &sa) != 0) {
			perror("MPII: restore sigaction");
		}
		
		wait_all_req();
	}

	dbgout("Did check");

	// Mark the last active request as a end-of-chain candidate.
	if (last_req) {
		last_req->faultat = addr;
		last_req = NULL;
	}
	dbgout("Sigact ret");
}

/*! \brief Initialize injection.
 *
 * The guts of this function will differ betweenthe static and dynamic
 * injection types, but their function is to initialize Orig_MPI_Init
 * and related functions.
 */

#ifdef INJECT_DYNAMIC
static void Inj_init_chain()
{
    // Use RTLD_NEXT to resolve in "the next library down the line"
    // which is either the original MPI library or other MPI tracer
    // code. This way we're as friendly as possible.
    Orig_MPI_Init = (int (*)(int*, char***))dlsym(RTLD_NEXT, "MPI_Init");
    Orig_MPI_Isend = (int (*)(void*, int, unsigned int, int, int, unsigned int, MPI_Request*))dlsym(RTLD_NEXT, "MPI_Isend");
    Orig_MPI_Irecv = (int (*)(void*, int, unsigned int, int, int, unsigned int, MPI_Request*))dlsym(RTLD_NEXT, "MPI_Irecv");
    Orig_MPI_Wait = (int (*)(MPI_Request*,  MPI_Status*))dlsym(RTLD_NEXT, "MPI_Wait");
    Orig_MPI_Get_count = (int (*)(MPI_Status *, MPI_Datatype, int *))dlsym(RTLD_NEXT, "MPI_Get_count");
    if (!Orig_MPI_Init) {
	fprintf(stderr, "MPII failed to find original MPI_Init\n");
	abort();
    } else {
	fprintf(stderr, "MPII .so injection\n");
    }
}
#else
static void Inj_init_chain()
{
    // Simply resolve the symbols at link time.
    Orig_MPI_Init = MPI_Init;
    Orig_MPI_Isend = MPI_Isend;
    Orig_MPI_Irecv = MPI_Irecv;
    Orig_MPI_Wait = MPI_Wait;
    Orig_MPI_Get_count = MPI_Get_count;
    Orig_malloc = malloc;
    Orig_free = free;
}
#endif

/*! \brief Initialize MPI and injection.
 * 
 * This is the overridden versionof MPI_Init(), and will
 * initialize the signal handler and status variables
 * in addition to calling the original MPI_Init().
 */

INJ_METHOD int Inj_MPI_Init(int *argc, char ***argv)
{

    if (getenv("INJ_DEBUG"))
	do_debug = atoi(getenv("INJ_DEBUG"));
    else
	do_debug = -1;

    if (getenv("INJ_IGNORE") && atoi(getenv("INJ_IGNORE")))
	do_nothing = 1;
    else
	do_nothing = 0;

    Inj_init_chain();


  sa.sa_sigaction = sigact;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_SIGINFO;


    int ret=Orig_MPI_Init(argc, argv);

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	MPI_Bcast(&do_debug, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&do_nothing, 1, MPI_INT, 0, MPI_COMM_WORLD);

  if (sigaction(SIGSEGV, &sa, &oldsa) != 0) {
    perror("MPII: sigaction");
  }

    dbgout("MPII: All is in order.");

    fprintf(stderr, "MPII Init (debug %d, ignore %d)\n", do_debug, do_nothing);

	return ret;
}

//! \brief Check if x is inside our overriden malloc()ed pages.
#define MEM_VALID_TEST(x) (mallocpages.find((unsigned long int)x / PAGE_SIZE) != mallocpages.end())

INJ_METHOD int Inj_MPI_Send(void *buf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm)
{
    if (do_nothing || !MEM_VALID_TEST(buf)) {
	    MPI_Request req;
	    MPI_Status stat;

	    Orig_MPI_Isend(buf, count, datatype, dest, tag, comm, &req);
	    return Orig_MPI_Wait(&req, &stat);
    }

    dbgout("MPI_Send chained to");
	add_req(true, buf, count, datatype, dest, tag, comm, NULL);
	return MPI_SUCCESS;
}

INJ_METHOD int Inj_MPI_Recv(void *buf, int count, MPI_Datatype datatype, int source, int tag, MPI_Comm comm, MPI_Status * status)
{
    if (do_nothing || !MEM_VALID_TEST(buf)) {
	    MPI_Request req;
	    Orig_MPI_Irecv(buf, count, datatype, source, tag, comm, &req);
	    return Orig_MPI_Wait(&req, status);
    }

    dbgout("MPI_Recv chained to");
    add_req(false, buf, count, datatype, source, tag, comm, status);
    return MPI_SUCCESS;
}

/*! \brief Overridden MPI_Sendrecv()
 *
 * If do_nothing is true, this just transforms the request into a Irecv and
 * Isend, but waits on them immediately before returning.
 *
 * During normal operation, add_req() is called for both the send and
 * receive parts of the request before we return.
 */
INJ_METHOD int Inj_MPI_Sendrecv(void *sendbuf, int sendcount, MPI_Datatype sendtype, int dest, int sendtag, void *recvbuf, int recvcount, MPI_Datatype recvtype, int source, int recvtag, MPI_Comm comm, MPI_Status * status)
{
    if (do_nothing || !MEM_VALID_TEST(sendbuf) || !MEM_VALID_TEST(recvbuf)) {
	    MPI_Request req[2];
    	MPI_Status stat[2];
	Orig_MPI_Irecv(recvbuf, recvcount, recvtype, source, recvtag, comm, &req[0]);	
	Orig_MPI_Isend(sendbuf, sendcount, sendtype, dest, sendtag, comm, &req[1]);

	MPI_Wait(&req[1], &stat[1]);
	return MPI_Wait(&req[0], status);
    }

    dbgout("MPI_Sendrecv chained to");

    MemReq *sendreq = add_req(true, sendbuf, sendcount, sendtype, dest, sendtag, comm, NULL);
    add_req(false, recvbuf, recvcount, recvtype, source, recvtag, comm, status, sendreq);

    dbgout("MPI_Sendrecv chained out");

    return MPI_SUCCESS;
}

/*! \brief Overridden MPI_Get_count()
 *
 * This waits for the request to actually finish, then chains through.
 */

INJ_METHOD int Inj_MPI_Get_count(MPI_Status *status, MPI_Datatype datatype, int *count) {
	dbgout("MPI_Get_count chained");

	map<MPI_Status *, MemReq *>::const_iterator iter;

	iter=statuses.find(status);
	if (iter != statuses.end())
		wait_req((*iter).second);

	return Orig_MPI_Get_count(status, datatype, count);
}

/*! \brief Overridden malloc().
 *
 * This is an overridden malloc() which ensures the allocated area is
 * to page boundries. Without this, small dynamic allocations (such as those
 * done by the MPI library itself) could end up in the same page as the
 * original buffer, meaning they too would be inaccessible once we
 * mprotect().
 *
 * However, if MPI_Init hasn't been called yet, this just chains through
 * immediately. If called via LD_PRELOAD, we would also be injected
 * into mpirun, and it doesn't make sense to pagealign that.
 */
INJ_METHOD void *Inj_malloc(size_t size) {
	if (! Orig_malloc) 
		    Orig_malloc = (void *(*)(size_t)) dlsym(RTLD_NEXT, "malloc");
	if (! Orig_MPI_Init)
		return Orig_malloc(size);

	void *base = Orig_malloc(size + 2*PAGE_SIZE);
	unsigned long int adjusted = ((long int)base + PAGE_SIZE - 1) & PAGE_MASK;
	mallocmap[(void *)adjusted] = base;
	mallocsizes[(void *)adjusted] = size;

	for(unsigned long int i=adjusted / PAGE_SIZE;i<=(adjusted+size)/PAGE_SIZE;i++)
		mallocpages.insert(i);

	return (void *) adjusted;
}

/*! \brief Overridden free().
 *
 * This is an overridden free() which replaces the pointer
 * with the "orignal" if this memory slab was allocated
 * through our overridden Inj_malloc().
 */

INJ_METHOD void Inj_free(void *ptr) {
	if (! Orig_free)
	    Orig_free = (void (*)(void *)) dlsym(RTLD_NEXT, "free");
	if (! Orig_MPI_Init)
		return Orig_free(ptr);

	void *base=mallocmap[ptr];
	if (base) {
		Orig_free(base);
		unsigned long int adjusted=(long int)ptr;
		for(unsigned long int i=adjusted / PAGE_SIZE;i<=(adjusted+mallocsizes[ptr])/PAGE_SIZE;i++)
			mallocpages.erase(i);
		mallocsizes.erase(ptr);
	}
	else
		Orig_free(ptr);
	mallocmap.erase(ptr);
}
