#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include <string.h>

// Static injection test
#include "mpii.h"

int rank;
int nump;
bool show = false;

#define DO_SIMPLE
#undef DO_ISEND
#undef DO_OVERLAP


void printnum(double v, double min=-1.0, double max=1.0) {
  if (v < min)
    v = min;
  if (v > max)
    v = max;
  double r = (v-min)/(max-min);

  bool bold = false;
  char c = 0;

  // Mrkbl, bl, mrk cyan, cyan, gul, mrk gul, rd, mrk rd, hvit

  if (v == 0.0) {
  } else if (r == 0.0) {
    c = 30;
    bold = true;
  } else if (r == 1.0) {
    c = 37;
    bold = true;
  } else if (r < 0.125) {
    c = 34;
  } else if (r < 0.250) {
    c = 34;
    bold = true;
  } else if (r < 0.375) {
    c = 36;
  } else if (r < 0.5) {
    c = 36;
    bold = true;
  } else if (r < 0.625) {
    c = 33;
    bold = true;
  } else if (r < 0.750) {
    c = 33;
  } else if (r < 0.875) {
    c = 31;
    bold = true;
  } else {
    c = 31;
  }

  fprintf(stderr,"%c[%dm", 27, c);
  if (bold)
    fprintf(stderr,"%c[1m",27);
  fprintf(stderr,"%6.3f ", v);
  fprintf(stderr,"%c[0m", 27);
}

struct Vector {
  friend class Matrix;
protected:
  double *vector;
  int dim;
public:
  Vector(int dim) {
    this->dim=dim;
    vector=(double *) malloc(dim * sizeof(double));
    memset(vector, sizeof(double) * dim, 0);
  }
  ~Vector() {
    free(vector);
  }
  operator double *() {
    return vector;
  }
  operator const double *() const {
    return vector;
  }
};

struct Matrix {
protected:
  Vector v;
  int _n, _m;
public:
  Matrix(int m, int n) : v(n*m) {
    _n = n;
    _m = m;
  }
  int n() const {
    return _n;
  }
  int m() const {
    return _m;
  }
  
  operator double *() {
    return v;
  }
  double *operator[](int i) {
    return v+_n*i;
  }
  const double * operator [](int i) const {
    return v+_n*i;
  }
};

/*
 * Mode of operation:
 * If f == NULL, assume the case where f(x,y) = 1
 * If u == NULL, return umax
 * If u != NULL, compare the estimated u with the
 * provided u and report emaxx
 */

typedef double (*PoissonFunc) (double, double);

double fone(double, double) {
  return 1.0L;
}

#define UPDATE(j,i) u[j][i]=(s/4.0) * (u[j-1][i]+u[j+1][i]+u[j][i-1]+u[j][i+1]) + (1.0 - s) * u[j][i]

void sorsolve(int n, int steps = 10, PoissonFunc ffunc = NULL, PoissonFunc ufunc = NULL) 
{
  double h;
  int i, j, t;
  double iv, jv;

  const double s= 1.8;

  Matrix u(n,n);

  h    = 1.0L/(double)(n-1);

  if (ffunc == NULL)
    ffunc = fone;

  for (j=0; j < n; j++) {
    for (i=0; i < n; i++) {
      iv = (double) (i) * h;
      jv = (double) (j) * h;
      u[j][i] = ffunc(iv,jv);
    }
  }

  for(t=1;t<=steps;t++) {
    printf("Iter %d (omega = %f)\n", t, s);

    for(j=1;j< n-1; j++) {
      for(i=(j & 1) ? 1 : 2;i<n-1;i+=2) {
        UPDATE(j, i);
      }
    }
    for(j=1;j< n-1; j++) {
      for(i=(j & 1) ? 2 : 1;i<n-1;i+=2) {
        UPDATE(j, i);
      }
    }

    for(j=0;j< n; j++) {
      for(i=0;i<n;i++) {
        printnum(u[j][i]);
      }
      printf("\n");
    }
  }
}

void reptiming(const int steps, double *elapsed, double *com_elapsed, const char *what) {
  double *min, *max, *avg, *com_min, *com_max, *com_avg;
  min = new double[steps];
  max = new double[steps];
  avg = new double[steps];
  com_min = new double[steps];
  com_max = new double[steps];
  com_avg = new double[steps];
  int i;

  MPI_Allreduce(elapsed, min, steps, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  MPI_Allreduce(elapsed, max, steps, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
  MPI_Allreduce(elapsed, avg, steps, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  for(i=0;i<steps;i++)
	avg[i] /= nump;

  MPI_Allreduce(com_elapsed, com_min, steps, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  MPI_Allreduce(com_elapsed, com_max, steps, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
  MPI_Allreduce(com_elapsed, com_avg, steps, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  for(i=0;i<steps;i++)
	com_avg[i] /= nump;

  if (rank == 0) {
    printf("%s\n",what);
    for(i=0;i<steps;i++)
	printf("Iter %3d %f -> %f [%f] (%fcpu %fcom)\n", i, min[i], max[i], avg[i], avg[i]-com_avg[i],com_avg[i]);
  }

  delete [] min;
  delete [] max;
  delete [] avg;
  delete [] com_min;
  delete [] com_max;
  delete [] com_avg;
}

// Upper left is red

void parasorsolve(const int n, const int steps = 10, PoissonFunc ffunc = NULL, PoissonFunc ufunc = NULL) 
{
  double h;
  int lx, ly;
  int i, j, t;
  double iv, jv;
  double umax;
  const double s= 1.4;

  const int showrank=show ? ((nump==1) ? 0 : nump-1) : -1;

  int dims[2];
  int periods[2] = {0, 0};
  MPI_Comm comm_cart;
  MPI_Group group_cart;
  int rank_cart;
  int coord_cart[2];
  MPI_Datatype type_column, type_row;
  MPI_Status status;
  int rank_left, rank_right, rank_up, rank_down;

  double start, stop, com_start, com_stop;

  double *elapsed = new double[steps];
  double *com_elapsed = new double[steps];

  dims[0] = dims[1] = 0;
  MPI_Dims_create(nump, 2, dims);

  if ((n % dims[0]) || (n % dims[1])) {
    if (rank==0) 
      printf("Failed to divide %dx%d grid into %dx%d nodes\n", n, n, dims[0],dims[1]);
    return;
  }

  lx = n / dims[0];
  ly = n / dims[1];

  if ((lx %2) || (ly % 2)) {
    if (rank==0)
	printf("Can't create datatypes with uneven red/black lengths\n");
    return;
  }


  if (rank == 0) 
    printf("Organized into %dx%d nodes (%dx%d points, %dx%d total)\n",dims[0],dims[1],lx,ly,n,n);

  MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, true, &comm_cart);
  MPI_Comm_group(comm_cart, &group_cart);
  MPI_Group_rank(group_cart, &rank_cart);
  MPI_Cart_coords(comm_cart, rank_cart, 2, coord_cart);

  MPI_Type_vector(ly / 2, 1, (lx+2) * 2, MPI_DOUBLE, &type_column);
  MPI_Type_commit(&type_column);

  MPI_Type_vector(lx / 2, 1, 2, MPI_DOUBLE, &type_row);
  MPI_Type_commit(&type_row);

  MPI_Cart_shift(comm_cart, 0, 1, &rank_left, &rank_right);
  MPI_Cart_shift(comm_cart, 1, 1, &rank_up, &rank_down);

  Matrix u(ly+2,lx+2);

  h    = 1.0L/(double)(n+1);

  if (ffunc == NULL)
    ffunc = fone;

  for (j=0; j <= ly+1; j++) {
    for (i=0; i <= lx+1; i++) {
      iv = (double) (i+coord_cart[0]*lx) * h;
      jv = (double) (j+coord_cart[1]*ly) * h;
      u[j][i] = ffunc(iv,jv);
    }
  }

#ifdef DO_SIMPLE
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Barrier(comm_cart);

  for(t=0;t<steps;t++) {
    if (rank==showrank) {
		fflush(stdout);
		fflush(stderr);    
	    fprintf(stderr,"Iter %d\n", t);
	}

    com_elapsed[t] = 0.0;
    start = MPI_Wtime();

    // Exchange black borders

    com_start = MPI_Wtime();

    MPI_Sendrecv(&u[1][2], 1, type_row, rank_up, 0, &u[ly+1][2], 1, type_row, rank_down, 0, comm_cart, &status);    
    MPI_Sendrecv(&u[ly][1], 1, type_row, rank_down, 0, &u[0][1], 1, type_row, rank_up, 0, comm_cart, &status);

    MPI_Sendrecv(&u[2][1], 1, type_column, rank_left, 0, &u[2][lx+1], 1, type_column, rank_right, 0, comm_cart, &status);
    MPI_Sendrecv(&u[1][lx], 1, type_column, rank_right, 0, &u[1][0], 1, type_column, rank_left, 0, comm_cart, &status);

    com_stop = MPI_Wtime();
    com_elapsed[t] += com_stop - com_start;

    // Compute red
    for(j=1;j<=ly; j++) {
      for(i=(j&1)?1:2;i<=lx;i+=2) {
        UPDATE(j, i);
      }
    }

    // Exchange red borders

    com_start = MPI_Wtime();

    MPI_Sendrecv(&u[1][1], 1, type_row, rank_up, 0, &u[ly+1][1], 1, type_row, rank_down, 0, comm_cart, &status);    
    MPI_Sendrecv(&u[ly][2], 1, type_row, rank_down, 0, &u[0][2], 1, type_row, rank_up, 0, comm_cart, &status);

    MPI_Sendrecv(&u[1][1], 1, type_column, rank_left, 0, &u[1][lx+1], 1, type_column, rank_right, 0, comm_cart, &status);
    MPI_Sendrecv(&u[2][lx], 1, type_column, rank_right, 0, &u[2][0], 1, type_column, rank_left, 0, comm_cart, &status);

    com_stop = MPI_Wtime();
    com_elapsed[t] += com_stop - com_start;

    // Compute black
    for(j=1;j<=ly; j++) {
      for(i=(j&1)?2:1;i<=lx;i+=2) {
        UPDATE(j, i);
      }
    }

    stop = MPI_Wtime();
    elapsed[t] = stop-start;

    if (rank==showrank)
    for(j=0;j<=ly+1; j++) {
      for(i=0;i<=lx+1;i++) {
        printnum(u[j][i]);
      }
      fprintf(stderr,"\n");
    }
  }

  reptiming(steps, elapsed, com_elapsed, "Regular sendrecv");

#endif

  MPI_Request req[8];
  MPI_Status statuses[8];

#ifdef DO_ISEND

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Barrier(comm_cart);

  for(t=0;t<steps;t++) {
    if (rank==showrank)    
	    printf("Iter %d (ISend)\n", t);

    com_elapsed[t] = 0.0;
    start = MPI_Wtime();

    // Exchange black borders
  
    com_start = MPI_Wtime();

    MPI_Irecv(&u[ly+1][2], 1, type_row, rank_down, 0, comm_cart, &req[0]);
    MPI_Irecv(&u[0][1], 1, type_row, rank_up, 0, comm_cart, &req[1]);
    MPI_Irecv(&u[2][lx+1], 1, type_column, rank_right, 0, comm_cart, &req[2]);
    MPI_Irecv(&u[1][0], 1, type_column, rank_left, 0, comm_cart, &req[3]);

    MPI_Isend(&u[1][2], 1, type_row, rank_up, 0, comm_cart, &req[4]);
    MPI_Isend(&u[ly][1], 1, type_row, rank_down, 0, comm_cart, &req[5]);
    MPI_Isend(&u[2][1], 1, type_column, rank_left, 0, comm_cart, &req[6]);
    MPI_Isend(&u[1][lx], 1, type_column, rank_right, 0, comm_cart, &req[7]);

    MPI_Waitall(8, req, statuses);

    com_stop = MPI_Wtime();
    com_elapsed[t] += com_stop - com_start;

    // Compute red
    for(j=1;j<=ly; j++) {
      for(i=(j&1)?1:2;i<=lx;i+=2) {
        UPDATE(j, i);
      }
    }

    // Exchange red borders

    com_start = MPI_Wtime();

    MPI_Irecv(&u[ly+1][1], 1, type_row, rank_down, 0, comm_cart, &req[0]);
    MPI_Irecv(&u[0][2], 1, type_row, rank_up, 0, comm_cart, &req[1]);
    MPI_Irecv(&u[1][lx+1], 1, type_column, rank_right, 0, comm_cart, &req[2]);
    MPI_Irecv(&u[2][0], 1, type_column, rank_left, 0, comm_cart, &req[3]);

    MPI_Isend(&u[1][1], 1, type_row, rank_up, 0, comm_cart, &req[4]);
    MPI_Isend(&u[ly][2], 1, type_row, rank_down, 0, comm_cart, &req[5]);
    MPI_Isend(&u[1][1], 1, type_column, rank_left, 0, comm_cart, &req[6]);
    MPI_Isend(&u[2][lx], 1, type_column, rank_right, 0, comm_cart, &req[7]);

    MPI_Waitall(8, req, statuses);

    com_stop = MPI_Wtime();
    com_elapsed[t] += com_stop - com_start;

    // Compute black
    for(j=1;j<=ly; j++) {
      for(i=(j&1)?2:1;i<=lx;i+=2) {
        UPDATE(j, i);
      }
    }

    stop = MPI_Wtime();
    elapsed[t] = stop-start;


    if (rank==showrank)
    for(j=0;j<=ly+1; j++) {
      for(i=0;i<=lx+1;i++) {
        printnum(u[j][i]);
      }
      printf("\n");
    }
  }

  reptiming(steps, elapsed, com_elapsed, "ISend/IRecv");

#endif

#ifdef DO_OVERLAP

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Barrier(comm_cart);

  // Exchange black borders
  MPI_Irecv(&u[ly+1][2], 1, type_row, rank_down, 0, comm_cart, &req[0]);
  MPI_Irecv(&u[0][1], 1, type_row, rank_up, 0, comm_cart, &req[1]);
  MPI_Irecv(&u[2][lx+1], 1, type_column, rank_right, 0, comm_cart, &req[2]);
  MPI_Irecv(&u[1][0], 1, type_column, rank_left, 0, comm_cart, &req[3]);

  MPI_Isend(&u[1][2], 1, type_row, rank_up, 0, comm_cart, &req[4]);
  MPI_Isend(&u[ly][1], 1, type_row, rank_down, 0, comm_cart, &req[5]);
  MPI_Isend(&u[2][1], 1, type_column, rank_left, 0, comm_cart, &req[6]);
  MPI_Isend(&u[1][lx], 1, type_column, rank_right, 0, comm_cart, &req[7]);

  for(t=0;t<steps;t++) {
    if (rank==showrank)    
	    printf("Iter %d (ISend overlap)\n", t);

    com_elapsed[t] = 0.0;
    start = MPI_Wtime();

    // Wait for black borders
    com_start = MPI_Wtime();
    MPI_Waitall(8, req, statuses);
    com_stop = MPI_Wtime();
    com_elapsed[t] += com_stop - com_start;

    // Compute red borders
    for(i=1;i<=lx;i+=2) {
       UPDATE(1, i);
       UPDATE(ly, i+1);
    }

    for(j=2;j<ly;j+=2) {
       UPDATE(j+1, 1);
       UPDATE(j, lx);
    }

    // Exchange red borders
    com_start = MPI_Wtime();
    MPI_Irecv(&u[ly+1][1], 1, type_row, rank_down, 0, comm_cart, &req[0]);
    MPI_Irecv(&u[0][2], 1, type_row, rank_up, 0, comm_cart, &req[1]);
    MPI_Irecv(&u[1][lx+1], 1, type_column, rank_right, 0, comm_cart, &req[2]);
    MPI_Irecv(&u[2][0], 1, type_column, rank_left, 0, comm_cart, &req[3]);

    MPI_Isend(&u[1][1], 1, type_row, rank_up, 0, comm_cart, &req[4]);
    MPI_Isend(&u[ly][2], 1, type_row, rank_down, 0, comm_cart, &req[5]);
    MPI_Isend(&u[1][1], 1, type_column, rank_left, 0, comm_cart, &req[6]);
    MPI_Isend(&u[2][lx], 1, type_column, rank_right, 0, comm_cart, &req[7]);
    com_stop = MPI_Wtime();
    com_elapsed[t] += com_stop - com_start;

    // Compute red interior
    for(j=2;j<ly; j++) {
      for(i=(j&1)?3:2;i<lx;i+=2) {
        UPDATE(j, i);
      }
    }


    // Wait for red borders
    com_start = MPI_Wtime();
    MPI_Waitall(8, req, statuses);
    com_stop = MPI_Wtime();
    com_elapsed[t] += com_stop - com_start;

    // Compute black borders
    for(i=1;i<=lx;i+=2) {
       UPDATE(1, i+1);
       UPDATE(ly, i);
    }

    for(j=2;j<ly;j+=2) {
       UPDATE(j, 1);
       UPDATE(j+1, lx);
    }

    // Exchange black borders
    com_start = MPI_Wtime();
    MPI_Irecv(&u[ly+1][2], 1, type_row, rank_down, 0, comm_cart, &req[0]);
    MPI_Irecv(&u[0][1], 1, type_row, rank_up, 0, comm_cart, &req[1]);
    MPI_Irecv(&u[2][lx+1], 1, type_column, rank_right, 0, comm_cart, &req[2]);
    MPI_Irecv(&u[1][0], 1, type_column, rank_left, 0, comm_cart, &req[3]);

    MPI_Isend(&u[1][2], 1, type_row, rank_up, 0, comm_cart, &req[4]);
    MPI_Isend(&u[ly][1], 1, type_row, rank_down, 0, comm_cart, &req[5]);
    MPI_Isend(&u[2][1], 1, type_column, rank_left, 0, comm_cart, &req[6]);
    MPI_Isend(&u[1][lx], 1, type_column, rank_right, 0, comm_cart, &req[7]);
    com_stop = MPI_Wtime();
    com_elapsed[t] += com_stop - com_start;

    // Compute black interior
    for(j=2;j<ly; j++) {
      for(i=(j&1)?2:3;i<lx;i+=2) {
        UPDATE(j, i);
      }
    }

    stop = MPI_Wtime();
    elapsed[t] = stop-start;

    if (rank==showrank)
    for(j=0;j<=ly+1; j++) {
      for(i=0;i<=lx+1;i++) {
        printnum(u[j][i]);
      }
      printf("\n");
    }
  }

  reptiming(steps, elapsed, com_elapsed, "ISend/IRecv overlapped");

#endif

  delete [] elapsed;
  delete [] com_elapsed;
}

double examplef(double x, double y) {
/*
  if (x == 0.5 && y == 0.5)
	return 1.0;
  return 0.0;
*/
  // Wiley coyote problem
  if (x == 0.0 || x == 1.0)
	return -1.0;
  if (y == 0.0)
	return -1.0;
  if (y == 1.0)
	return 1.0;
  return 0.0;
}

double exampleu(double x, double y) {
  return sin(M_PI * x) * sin(2.0L * M_PI * y);
}


main(int argc, char **argv )
{
  int n = 8;

  MPI_Init(&argc, &argv);
  
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &nump);

  if ((rank == 0) && (argc == 2))
    n = atoi(argv[1]);

  MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);

  if (n < 20)
	show = true;

//  sorsolve(n, 40, examplef, exampleu);

  parasorsolve(n, 20, examplef, exampleu);

  MPI_Finalize();
}
